@@ -4,9 +4,10 @@
 # See the COPYING file for more information.
 
 import xml.etree.ElementTree as ET
-import string
-import StringIO
 import logging
+
+from musicbrainzngs import util
+
 try:
 	from ET import fixtag
 except:
@@ -16,7 +17,7 @@ except:
 		# tag and namespace declaration, if any
 		if isinstance(tag, ET.QName):
 			tag = tag.text
-		namespace_uri, tag = string.split(tag[1:], "}", 1)
+		namespace_uri, tag = tag[1:].split("}", 1)
 		prefix = namespaces.get(namespace_uri)
 		if prefix is None:
 			prefix = "ns%d" % len(namespaces)
@@ -29,14 +30,20 @@ except:
 			xmlns = None
 		return "%s:%s" % (prefix, tag), xmlns
 
-NS_MAP = {"http://musicbrainz.org/ns/mmd-2.0#": "ws2"}
+
+NS_MAP = {"http://musicbrainz.org/ns/mmd-2.0#": "ws2",
+          "http://musicbrainz.org/ns/ext#-2.0": "ext"}
 _log = logging.getLogger("python-musicbrainz-ngs")
 
 def make_artist_credit(artists):
 	names = []
 	for artist in artists:
 		if isinstance(artist, dict):
-			names.append(artist.get("artist", {}).get("name", ""))
+			if "name" in artist:
+				names.append(artist.get("name", ""))
+			else:
+				names.append(artist.get("artist", {}).get("name", ""))
+
 		else:
 			names.append(artist)
 	return "".join(names)
@@ -56,25 +63,30 @@ def parse_elements(valid_els, element):
 		if ":" in t:
 			t = t.split(":")[1]
 		if t in valid_els:
-			result[t] = sub.text
+			result[t] = sub.text or ""
 		else:
 			_log.debug("in <%s>, uncaught <%s>", fixtag(element.tag, NS_MAP)[0], t)
 	return result
 
 def parse_attributes(attributes, element):
-	""" Extract attributes from an element.
-	    For example, given the element:
-	    <element type="Group" />
-	    and a list attributes that contains "type",
-	    return a dict {'type': 'Group'}
-	"""
-	result = {}
-	for attr in attributes:
-		if attr in element.attrib:
-			result[attr] = element.attrib[attr]
-		else:
-			_log.debug("in <%s>, uncaught attribute %s", fixtag(element.tag, NS_MAP)[0], attr)
-	return result
+    """ Extract attributes from an element.
+        For example, given the element:
+        <element type="Group" />
+        and a list attributes that contains "type",
+        return a dict {'type': 'Group'}
+    """
+    result = {}
+    for attr in element.attrib:
+        if "{" in attr:
+            a = fixtag(attr, NS_MAP)[0]
+        else:
+            a = attr
+        if a in attributes:
+            result[a] = element.attrib[attr]
+        else:
+            _log.debug("in <%s>, uncaught attribute %s", fixtag(element.tag, NS_MAP)[0], attr)
+
+    return result
 
 def parse_inner(inner_els, element):
 	""" Delegate the parsing of a subelement to another function.
@@ -107,9 +119,7 @@ def parse_inner(inner_els, element):
 	return result
 
 def parse_message(message):
-	s = message.read()
-	f = StringIO.StringIO(s)
-	tree = ET.ElementTree(file=f)
+	tree = util.bytes_to_elementtree(message)
 	root = tree.getroot()
 	result = {}
 	valid_elements = {"artist": parse_artist,
@@ -123,13 +133,14 @@ def parse_message(message):
 	                  "puid": parse_puid,
 	                  "echoprint": parse_puid,
 
+	                  "annotation-list": parse_annotation_list,
 	                  "artist-list": parse_artist_list,
 	                  "label-list": parse_label_list,
 	                  "release-list": parse_release_list,
 	                  "release-group-list": parse_release_group_list,
 	                  "recording-list": parse_recording_list,
 	                  "work-list": parse_work_list,
-	
+
 	                  "collection-list": parse_collection_list,
 	                  "collection": parse_collection,
 
@@ -159,6 +170,17 @@ def parse_collection_release_list(rl):
 	attribs = ["count"]
 	return parse_attributes(attribs, rl)
 
+def parse_annotation_list(al):
+	return [parse_annotation(a) for a in al]
+
+def parse_annotation(annotation):
+	result = {}
+	attribs = ["type", "ext:score"]
+	elements = ["entity", "name", "text"]
+	result.update(parse_attributes(attribs, annotation))
+	result.update(parse_elements(elements, annotation))
+	return result
+
 def parse_artist_lifespan(lifespan):
 	parts = parse_elements(["begin", "end"], lifespan)
 
@@ -168,50 +190,51 @@ def parse_artist_list(al):
 	return [parse_artist(a) for a in al]
 
 def parse_artist(artist):
-	result = {}
-	attribs = ["id", "type"]
-	elements = ["name", "sort-name", "country", "user-rating"]
-	inner_els = {"life-span": parse_artist_lifespan,
-	             "recording-list": parse_recording_list,
-	             "release-list": parse_release_list,
-	             "release-group-list": parse_release_group_list,
-	             "work-list": parse_work_list,
-	             "tag-list": parse_tag_list,
-	             "user-tag-list": parse_tag_list,
-	             "rating": parse_rating,
-	             "alias-list": parse_alias_list}
+    result = {}
+    attribs = ["id", "type", "ext:score"]
+    elements = ["name", "sort-name", "country", "user-rating",
+                "disambiguation", "gender", "ipi"]
+    inner_els = {"life-span": parse_artist_lifespan,
+                 "recording-list": parse_recording_list,
+                 "relation-list": parse_relation_list,
+                 "release-list": parse_release_list,
+                 "release-group-list": parse_release_group_list,
+                 "work-list": parse_work_list,
+                 "tag-list": parse_tag_list,
+                 "user-tag-list": parse_tag_list,
+                 "rating": parse_rating,
+                 "ipi-list": parse_element_list,
+                 "alias-list": parse_alias_list,
+                 "annotation": parse_annotation}
 
-	result.update(parse_attributes(attribs, artist))
-	result.update(parse_elements(elements, artist))
-	result.update(parse_inner(inner_els, artist))
+    result.update(parse_attributes(attribs, artist))
+    result.update(parse_elements(elements, artist))
+    result.update(parse_inner(inner_els, artist))
 
-	return result
+    return result
 
 def parse_label_list(ll):
-	return [parse_label(l) for l in ll]
+    return [parse_label(l) for l in ll]
 
 def parse_label(label):
-	result = {}
-	attribs = ["id", "type"]
-	elements = ["name", "sort-name", "country", "label-code", "user-rating"]
-	inner_els = {"life-span": parse_artist_lifespan,
-	             "release-list": parse_release_list,
-	             "tag-list": parse_tag_list,
-	             "user-tag-list": parse_tag_list,
-	             "rating": parse_rating,
-	             "alias-list": parse_alias_list}
+    result = {}
+    attribs = ["id", "type", "ext:score"]
+    elements = ["name", "sort-name", "country", "label-code", "user-rating",
+                "ipi", "disambiguation"]
+    inner_els = {"life-span": parse_artist_lifespan,
+                 "release-list": parse_release_list,
+                 "tag-list": parse_tag_list,
+                 "user-tag-list": parse_tag_list,
+                 "rating": parse_rating,
+                 "ipi-list": parse_element_list,
+                 "alias-list": parse_alias_list,
+                 "annotation": parse_annotation}
 
-	result.update(parse_attributes(attribs, label))
-	result.update(parse_elements(elements, label))
-	result.update(parse_inner(inner_els, label))
+    result.update(parse_attributes(attribs, label))
+    result.update(parse_elements(elements, label))
+    result.update(parse_inner(inner_els, label))
 
-	return result
-
-def parse_attribute_list(al):
-    return [parse_attribute_tag(a) for a in al]
-
-def parse_attribute_tag(attribute):
-    return attribute.text
+    return result
 
 def parse_relation_list(rl):
     attribs = ["target-type"]
@@ -228,7 +251,7 @@ def parse_relation(relation):
                  "recording": parse_recording,
                  "release": parse_release,
                  "release-group": parse_release_group,
-                 "attribute-list": parse_attribute_list,
+                 "attribute-list": parse_element_list,
                  "work": parse_work
                 }
     result.update(parse_attributes(attribs, relation))
@@ -239,14 +262,15 @@ def parse_relation(relation):
 
 def parse_release(release):
 	result = {}
-	attribs = ["id"]
+	attribs = ["id", "ext:score"]
 	elements = ["title", "status", "disambiguation", "quality", "country", "barcode", "date", "packaging", "asin"]
 	inner_els = {"text-representation": parse_text_representation,
 	             "artist-credit": parse_artist_credit,
 	             "label-info-list": parse_label_info_list,
 	             "medium-list": parse_medium_list,
 	             "release-group": parse_release_group,
-	             "relation-list": parse_relation_list}
+	             "relation-list": parse_relation_list,
+	             "annotation": parse_annotation}
 
 	result.update(parse_attributes(attribs, release))
 	result.update(parse_elements(elements, release))
@@ -276,26 +300,28 @@ def parse_text_representation(textr):
 	return parse_elements(["language", "script"], textr)
 
 def parse_release_group(rg):
-	result = {}
-	attribs = ["id", "type"]
-	elements = ["title", "user-rating", "first-release-date"]
-	inner_els = {"artist-credit": parse_artist_credit,
-	             "release-list": parse_release_list,
-	             "tag-list": parse_tag_list,
-	             "user-tag-list": parse_tag_list,
-	             "rating": parse_rating}
+    result = {}
+    attribs = ["id", "type", "ext:score"]
+    elements = ["title", "user-rating", "first-release-date", "primary-type"]
+    inner_els = {"artist-credit": parse_artist_credit,
+                 "release-list": parse_release_list,
+                 "tag-list": parse_tag_list,
+                 "user-tag-list": parse_tag_list,
+                 "secondary-type-list": parse_element_list,
+                 "rating": parse_rating,
+                 "annotation": parse_annotation}
 
-	result.update(parse_attributes(attribs, rg))
-	result.update(parse_elements(elements, rg))
-	result.update(parse_inner(inner_els, rg))
-	if "artist-credit" in result:
-		result["artist-credit-phrase"] = make_artist_credit(result["artist-credit"])
+    result.update(parse_attributes(attribs, rg))
+    result.update(parse_elements(elements, rg))
+    result.update(parse_inner(inner_els, rg))
+    if "artist-credit" in result:
+        result["artist-credit-phrase"] = make_artist_credit(result["artist-credit"])
 
-	return result
+    return result
 
 def parse_recording(recording):
 	result = {}
-	attribs = ["id"]
+	attribs = ["id", "ext:score"]
 	elements = ["title", "length", "user-rating"]
 	inner_els = {"artist-credit": parse_artist_credit,
 	             "release-list": parse_release_list,
@@ -304,7 +330,9 @@ def parse_recording(recording):
 	             "rating": parse_rating,
 	             "puid-list": parse_external_id_list,
 	             "isrc-list": parse_external_id_list,
-	             "echoprint-list": parse_external_id_list}
+	             "echoprint-list": parse_external_id_list,
+	             "relation-list": parse_relation_list,
+	             "annotation": parse_annotation}
 
 	result.update(parse_attributes(attribs, recording))
 	result.update(parse_elements(elements, recording))
@@ -317,26 +345,29 @@ def parse_recording(recording):
 def parse_external_id_list(pl):
 	return [parse_attributes(["id"], p)["id"] for p in pl]
 
+def parse_element_list(el):
+    return [e.text for e in el]
+
 def parse_work_list(wl):
-	result = []
-	for w in wl:
-		result.append(parse_work(w))
-	return result
+    return [parse_work(w) for w in wl]
 
 def parse_work(work):
-	result = {}
-	attribs = ["id"]
-	elements = ["title", "user-rating"]
-	inner_els = {"tag-list": parse_tag_list,
-	             "user-tag-list": parse_tag_list,
-	             "rating": parse_rating,
-	             "alias-list": parse_alias_list}
+    result = {}
+    attribs = ["id", "ext:score"]
+    elements = ["title", "user-rating", "language", "iswc"]
+    inner_els = {"tag-list": parse_tag_list,
+                 "user-tag-list": parse_tag_list,
+                 "rating": parse_rating,
+                 "alias-list": parse_alias_list,
+                 "iswc-list": parse_element_list,
+                 "relation-list": parse_relation_list,
+                 "annotation": parse_response_message}
 
-	result.update(parse_attributes(attribs, work))
-	result.update(parse_elements(elements, work))
-	result.update(parse_inner(inner_els, work))
+    result.update(parse_attributes(attribs, work))
+    result.update(parse_elements(elements, work))
+    result.update(parse_inner(inner_els, work))
 
-	return result
+    return result
 
 def parse_disc(disc):
 	result = {}
@@ -420,19 +451,29 @@ def parse_track_list(tl):
 	return result
 
 def parse_track(track):
-	result = {}
-	elements = ["position", "title"]
-	inner_els = {"recording": parse_recording}
+    result = {}
+    elements = ["number", "position", "title", "length"]
+    inner_els = {"recording": parse_recording,
+                 "artist-credit": parse_artist_credit}
 
-	result.update(parse_elements(elements, track))
-	result.update(parse_inner(inner_els, track))
-	return result
+    result.update(parse_elements(elements, track))
+    result.update(parse_inner(inner_els, track))
+    if "artist-credit" in result.get("recording", {}) and "artist-credit" not in result:
+        result["artist-credit"] = result["recording"]["artist-credit"]
+    if "artist-credit" in result:
+        result["artist-credit-phrase"] = make_artist_credit(result["artist-credit"])
+    # Make a length field that contains track length or recording length
+    track_or_recording = None
+    if "length" in result:
+        track_or_recording = result["length"]
+    elif result.get("recording", {}).get("length"):
+        track_or_recording = result.get("recording", {}).get("length")
+    if track_or_recording:
+        result["track_or_recording_length"] = track_or_recording
+    return result
 
 def parse_tag_list(tl):
-	result = []
-	for t in tl:
-		result.append(parse_tag(t))
-	return result
+    return [parse_tag(t) for t in tl]
 
 def parse_tag(tag):
 	result = {}
@@ -454,10 +495,16 @@ def parse_rating(rating):
 	return result
 
 def parse_alias_list(al):
-	result = []
-	for a in al:
-		result.append(a.text)
-	return result
+    return [parse_alias(a) for a in al]
+
+def parse_alias(alias):
+    result = {}
+    attribs = ["locale", "sort-name", "type", "primary", "begin-date", "end-date"]
+
+    result.update(parse_attributes(attribs, alias))
+    result["alias"] = alias.text
+
+    return result
 
 ###
 def make_barcode_request(barcodes):
@@ -541,9 +588,24 @@ def make_rating_request(artist_ratings, recording_ratings):
 	for art, rating in artist_ratings.items():
 		art_xml = ET.SubElement(art_list, "{%s}artist" % NS)
 		art_xml.set("{%s}id" % NS, art)
-		rating_xml = ET.SubElement(rec_xml, "{%s}user-rating" % NS)
+		rating_xml = ET.SubElement(art_xml, "{%s}user-rating" % NS)
 		if isinstance(rating, int):
 			rating = "%d" % rating
 		rating_xml.text = rating
 
 	return ET.tostring(root, "utf-8")
+
+def make_isrc_request(recordings_isrcs):
+	NS = "http://musicbrainz.org/ns/mmd-2.0#"
+	root = ET.Element("{%s}metadata" % NS)
+	rec_list = ET.SubElement(root, "{%s}recording-list" % NS)
+	for rec, isrcs in recordings_isrcs.items():
+		if len(isrcs) > 0:
+			rec_xml = ET.SubElement(rec_list, "{%s}recording" % NS)
+			rec_xml.set("{%s}id" % NS, rec)
+			isrc_list_xml = ET.SubElement(rec_xml, "{%s}isrc-list" % NS)
+			isrc_list_xml.set("{%s}count" % NS, str(len(isrcs)))
+			for isrc in isrcs:
+				isrc_xml = ET.SubElement(isrc_list_xml, "{%s}isrc" % NS)
+				isrc_xml.set("{%s}id" % NS, isrc)
+	return ET.tostring(root, "utf-8")
