@@ -3,20 +3,19 @@
 # This file is distributed under a BSD-2-Clause type license.
 # See the COPYING file for more information.
 
-import urlparse
-import urllib2
-import urllib
-import mbxml
 import re
 import threading
 import time
 import logging
-import httplib
 import socket
 import xml.etree.ElementTree as etree
 from xml.parsers import expat
 
-_version = "0.2dev"
+from musicbrainzngs import mbxml
+from musicbrainzngs import util
+from musicbrainzngs import compat
+
+_version = "0.3"
 _log = logging.getLogger("musicbrainzngs")
 
 
@@ -28,50 +27,69 @@ VALID_INCLUDES = {
 		"various-artists", "discids", "media",
 		"aliases", "tags", "user-tags", "ratings", "user-ratings", # misc
 		"artist-rels", "label-rels", "recording-rels", "release-rels",
-		"release-group-rels", "url-rels", "work-rels"
+		"release-group-rels", "url-rels", "work-rels", "annotation"
+	],
+	'annotation': [
+
 	],
 	'label': [
 		"releases", # Subqueries
 	    "discids", "media",
 	    "aliases", "tags", "user-tags", "ratings", "user-ratings", # misc
 		"artist-rels", "label-rels", "recording-rels", "release-rels",
-		"release-group-rels", "url-rels", "work-rels"
+		"release-group-rels", "url-rels", "work-rels", "annotation"
 	],
 	'recording': [
 		"artists", "releases", # Subqueries
 	    "discids", "media", "artist-credits",
 	    "tags", "user-tags", "ratings", "user-ratings", # misc
 		"artist-rels", "label-rels", "recording-rels", "release-rels",
-		"release-group-rels", "url-rels", "work-rels"
+		"release-group-rels", "url-rels", "work-rels", "annotation", "aliases"
 	],
 	'release': [
 		"artists", "labels", "recordings", "release-groups", "media",
 		"artist-credits", "discids", "puids", "echoprints", "isrcs",
 		"artist-rels", "label-rels", "recording-rels", "release-rels",
 		"release-group-rels", "url-rels", "work-rels", "recording-level-rels",
-		"work-level-rels"
+		"work-level-rels", "annotation", "aliases"
 	],
 	'release-group': [
 		"artists", "releases", "discids", "media",
 		"artist-credits", "tags", "user-tags", "ratings", "user-ratings", # misc
 		"artist-rels", "label-rels", "recording-rels", "release-rels",
-		"release-group-rels", "url-rels", "work-rels"
+		"release-group-rels", "url-rels", "work-rels", "annotation", "aliases"
 	],
 	'work': [
 		"artists", # Subqueries
 	    "aliases", "tags", "user-tags", "ratings", "user-ratings", # misc
 		"artist-rels", "label-rels", "recording-rels", "release-rels",
-		"release-group-rels", "url-rels", "work-rels"
+		"release-group-rels", "url-rels", "work-rels", "annotation"
 	],
 	'discid': [
-		"artists", "labels", "recordings", "release-groups", "puids",
-		"echoprints", "isrcs"
+		"artists", "labels", "recordings", "release-groups", "media",
+		"artist-credits", "discids", "puids", "echoprints", "isrcs",
+		"artist-rels", "label-rels", "recording-rels", "release-rels",
+		"release-group-rels", "url-rels", "work-rels", "recording-level-rels",
+		"work-level-rels"
 	],
 	'echoprint': ["artists", "releases"],
 	'puid': ["artists", "releases", "puids", "echoprints", "isrcs"],
 	'isrc': ["artists", "releases", "puids", "echoprints", "isrcs"],
 	'iswc': ["artists"],
+	'collection': ['releases'],
 }
+VALID_BROWSE_INCLUDES = {
+    'releases': ["artist-credits", "labels", "recordings",
+                "release-groups", "media", "discids", "artist-rels",
+                "label-rels", "recording-rels", "release-rels",
+                "release-group-rels", "url-rels", "work-rels"],
+    'recordings': ["artist-credits", "tags", "ratings", "user-tags",
+                  "user-ratings"],
+    'labels': ["aliases", "tags", "ratings", "user-tags", "user-ratings"],
+    'artists': ["aliases", "tags", "ratings", "user-tags", "user-ratings"],
+    'release-groups': ["artist-credits", "tags", "ratings", "user-tags", "user-ratings"]
+}
+
 VALID_RELEASE_TYPES = [
 	"nat", "album", "single", "ep", "compilation", "soundtrack", "spokenword",
 	"interview", "audiobook", "live", "remix", "other"
@@ -80,29 +98,36 @@ VALID_RELEASE_STATUSES = ["official", "promotion", "bootleg", "pseudo-release"]
 VALID_SEARCH_FIELDS = {
 	'artist': [
 		'arid', 'artist', 'sortname', 'type', 'begin', 'end', 'comment',
-		'alias', 'country', 'gender', 'tag'
+		'alias', 'country', 'gender', 'tag', 'ipi', 'artistaccent'
+	],
+	'annotation': [
+		'entity', 'name', 'text', 'type'
 	],
 	'release-group': [
 		'rgid', 'releasegroup', 'reid', 'release', 'arid', 'artist',
-		'artistname', 'creditname', 'type', 'tag'
+		'artistname', 'creditname', 'type', 'tag', 'releasegroupaccent',
+		'releases', 'comment'
 	],
 	'release': [
 		'reid', 'release', 'arid', 'artist', 'artistname', 'creditname',
 		'type', 'status', 'tracks', 'tracksmedium', 'discids',
 		'discidsmedium', 'mediums', 'date', 'asin', 'lang', 'script',
-		'country', 'date', 'label', 'catno', 'barcode', 'puid'
+		'country', 'date', 'label', 'catno', 'barcode', 'puid', 'comment',
+		'format', 'releaseaccent', 'rgid'
 	],
 	'recording': [
 		'rid', 'recording', 'isrc', 'arid', 'artist', 'artistname',
 		'creditname', 'reid', 'release', 'type', 'status', 'tracks',
-		'tracksrelease', 'dur', 'qdur', 'tnum', 'position', 'tag'
+		'tracksrelease', 'dur', 'qdur', 'tnum', 'position', 'tag', 'comment',
+		'country', 'date' 'format', 'recordingaccent'
 	],
 	'label': [
 		'laid', 'label', 'sortname', 'type', 'code', 'country', 'begin',
-		'end', 'comment', 'alias', 'tag'
+		'end', 'comment', 'alias', 'tag', 'ipi', 'labelaccent'
 	],
 	'work': [
-		'wid', 'work', 'iswc', 'type', 'arid', 'artist', 'alias', 'tag'
+		'wid', 'work', 'iswc', 'type', 'arid', 'artist', 'alias', 'tag',
+		'comment', 'workaccent'
 	],
 }
 
@@ -163,6 +188,10 @@ class ResponseError(WebServiceError):
 	"""Bad response sent by the MB server."""
 	pass
 
+class AuthenticationError(WebServiceError):
+	"""Received a HTTP 401 response while accessing a protected resource."""
+	pass
+
 
 # Helpers for validating and formatting allowed sets.
 
@@ -183,9 +212,9 @@ def _check_filter_and_make_params(entity, includes, release_status=[], release_t
 	the filters can be used with the given includes. Return a params
 	dict that can be passed to _do_mb_query.
 	"""
-	if isinstance(release_status, basestring):
+	if isinstance(release_status, compat.basestring):
 		release_status = [release_status]
-	if isinstance(release_type, basestring):
+	if isinstance(release_type, compat.basestring):
 		release_type = [release_type]
 	_check_filter(release_status, VALID_RELEASE_STATUSES)
 	_check_filter(release_type, VALID_RELEASE_TYPES)
@@ -223,9 +252,11 @@ def auth(u, p):
 	password = p
 
 def set_useragent(app, version, contact=None):
-    """ Set the User-Agent to be used for requests to the MusicBrainz webservice.
-    This should be set before requests are made."""
+    """Set the User-Agent to be used for requests to the MusicBrainz webservice.
+    This must be set before requests are made."""
     global _useragent, _client
+    if not app or not version:
+        raise ValueError("App and version can not be empty")
     if contact is not None:
         _useragent = "%s/%s python-musicbrainz-ngs/%s ( %s )" % (app, version, _version, contact)
     else:
@@ -234,76 +265,88 @@ def set_useragent(app, version, contact=None):
     _log.debug("set user-agent to %s" % _useragent)
 
 def set_hostname(new_hostname):
-	global hostname
-	hostname = new_hostname
+    """Set the base hostname for MusicBrainz webservice requests.
+    Defaults to 'musicbrainz.org'."""
+    global hostname
+    hostname = new_hostname
 
 # Rate limiting.
 
 limit_interval = 1.0
 limit_requests = 1
+do_rate_limit = True
 
-def set_rate_limit(new_interval=1.0, new_requests=1):
-	"""Sets the rate limiting behavior of the module. Must be invoked
-	before the first Web service call.  Specify the number of requests
-	(`new_requests`) that may be made per given interval
-	(`new_interval`).
-	"""
-	global limit_interval
-	global limit_requests
-	limit_interval = new_interval
-	limit_requests = new_requests
+def set_rate_limit(limit_or_interval=1.0, new_requests=1):
+    """Sets the rate limiting behavior of the module. Must be invoked
+    before the first Web service call.
+    If the `limit_or_interval` parameter is set to False then
+    rate limiting will be disabled. If it is a number then only
+    a set number of requests (`new_requests`) will be made per
+    given interval (`limit_or_interval`).
+    """
+    global limit_interval
+    global limit_requests
+    global do_rate_limit
+    if isinstance(limit_or_interval, bool):
+        do_rate_limit = limit_or_interval
+    else:
+        if limit_or_interval <= 0.0:
+            raise ValueError("limit_or_interval can't be less than 0")
+        if new_requests <= 0:
+            raise ValueError("new_requests can't be less than 0")
+        do_rate_limit = True
+        limit_interval = limit_or_interval
+        limit_requests = new_requests
 
 class _rate_limit(object):
-	"""A decorator that limits the rate at which the function may be
-	called. The rate is controlled by the `limit_interval` and
-	`limit_requests` global variables.  The limiting is thread-safe;
-	only one thread may be in the function at a time (acts like a
-	monitor in this sense). The globals must be set before the first
-	call to the limited function.
-	"""
-	def __init__(self, fun):
-		self.fun = fun
-		self.last_call = 0.0
-		self.lock = threading.Lock()
-		self.remaining_requests = None # Set on first invocation.
+    """A decorator that limits the rate at which the function may be
+    called. The rate is controlled by the `limit_interval` and
+    `limit_requests` global variables.  The limiting is thread-safe;
+    only one thread may be in the function at a time (acts like a
+    monitor in this sense). The globals must be set before the first
+    call to the limited function.
+    """
+    def __init__(self, fun):
+        self.fun = fun
+        self.last_call = 0.0
+        self.lock = threading.Lock()
+        self.remaining_requests = None # Set on first invocation.
 
-	def _update_remaining(self):
-		"""Update remaining requests based on the elapsed time since
-		they were last calculated.
-		"""
-		# On first invocation, we have the maximum number of requests
-		# available.
-		if self.remaining_requests is None:
-			self.remaining_requests = float(limit_requests)
+    def _update_remaining(self):
+        """Update remaining requests based on the elapsed time since
+        they were last calculated.
+        """
+        # On first invocation, we have the maximum number of requests
+        # available.
+        if self.remaining_requests is None:
+            self.remaining_requests = float(limit_requests)
 
-		else:
-			since_last_call = time.time() - self.last_call
-			self.remaining_requests += since_last_call * \
-									   (limit_requests / limit_interval)
-			self.remaining_requests = min(self.remaining_requests,
-										  float(limit_requests))
+        else:
+            since_last_call = time.time() - self.last_call
+            self.remaining_requests += since_last_call * \
+                                       (limit_requests / limit_interval)
+            self.remaining_requests = min(self.remaining_requests,
+                                          float(limit_requests))
 
-		self.last_call = time.time()
+        self.last_call = time.time()
 
-	def __call__(self, *args, **kwargs):
-		with self.lock:
-			self._update_remaining()
+    def __call__(self, *args, **kwargs):
+        with self.lock:
+            if do_rate_limit:
+                self._update_remaining()
 
-			# Delay if necessary.
-			while self.remaining_requests < 0.999:
-				time.sleep((1.0 - self.remaining_requests) *
-						   (limit_requests / limit_interval))
-				self._update_remaining()
+                # Delay if necessary.
+                while self.remaining_requests < 0.999:
+                    time.sleep((1.0 - self.remaining_requests) *
+                               (limit_requests / limit_interval))
+                    self._update_remaining()
 
-			# Call the original function, "paying" for this call.
-			self.remaining_requests -= 1.0
-			return self.fun(*args, **kwargs)
-
-
-# Generic support for making HTTP requests.
+                # Call the original function, "paying" for this call.
+                self.remaining_requests -= 1.0
+            return self.fun(*args, **kwargs)
 
 # From pymb2
-class _RedirectPasswordMgr(urllib2.HTTPPasswordMgr):
+class _RedirectPasswordMgr(compat.HTTPPasswordMgr):
 	def __init__(self):
 		self._realms = { }
 
@@ -318,18 +361,18 @@ class _RedirectPasswordMgr(urllib2.HTTPPasswordMgr):
 		# ignoring the uri parameter intentionally
 		self._realms[realm] = (username, password)
 
-class _DigestAuthHandler(urllib2.HTTPDigestAuthHandler):
+class _DigestAuthHandler(compat.HTTPDigestAuthHandler):
 	def get_authorization (self, req, chal):
 		qop = chal.get ('qop', None)
 		if qop and ',' in qop and 'auth' in qop.split (','):
 			chal['qop'] = 'auth'
 
-		return urllib2.HTTPDigestAuthHandler.get_authorization (self, req, chal)
+		return compat.HTTPDigestAuthHandler.get_authorization (self, req, chal)
 
-class _MusicbrainzHttpRequest(urllib2.Request):
+class _MusicbrainzHttpRequest(compat.Request):
 	""" A custom request handler that allows DELETE and PUT"""
 	def __init__(self, method, url, data=None):
-		urllib2.Request.__init__(self, url, data)
+		compat.Request.__init__(self, url, data)
 		allowed_m = ["GET", "POST", "DELETE", "PUT"]
 		if method not in allowed_m:
 			raise ValueError("invalid method: %s" % method)
@@ -341,11 +384,11 @@ class _MusicbrainzHttpRequest(urllib2.Request):
 
 # Core (internal) functions for calling the MB API.
 
-def _safe_open(opener, req, body=None, max_retries=8, retry_delay_delta=2.0):
+def _safe_read(opener, req, body=None, max_retries=8, retry_delay_delta=2.0):
 	"""Open an HTTP request with a given URL opener and (optionally) a
 	request body. Transient errors lead to retries.  Permanent errors
 	and repeated errors are translated into a small set of handleable
-	exceptions. Returns a file-like object.
+	exceptions. Return a bytestring.
 	"""
 	last_exc = None
 	for retry_num in range(max_retries):
@@ -358,36 +401,43 @@ def _safe_open(opener, req, body=None, max_retries=8, retry_delay_delta=2.0):
 				f = opener.open(req, body)
 			else:
 				f = opener.open(req)
+			return f.read()
 
-		except urllib2.HTTPError, exc:
-			if exc.code in (400, 404):
+		except compat.HTTPError as exc:
+			if exc.code in (400, 404, 411):
 				# Bad request, not found, etc.
 				raise ResponseError(cause=exc)
 			elif exc.code in (503, 502, 500):
 				# Rate limiting, internal overloading...
 				_log.debug("HTTP error %i" % exc.code)
+			elif exc.code in (401, ):
+				raise AuthenticationError(cause=exc)
 			else:
 				# Other, unknown error. Should handle more cases, but
 				# retrying for now.
 				_log.debug("unknown HTTP error %i" % exc.code)
 			last_exc = exc
-		except httplib.BadStatusLine, exc:
+		except compat.BadStatusLine as exc:
 			_log.debug("bad status line")
 			last_exc = exc
-		except httplib.HTTPException, exc:
+		except compat.HTTPException as exc:
 			_log.debug("miscellaneous HTTP exception: %s" % str(exc))
 			last_exc = exc
-		except urllib2.URLError, exc:
+		except compat.URLError as exc:
 			if isinstance(exc.reason, socket.error):
 				code = exc.reason.errno
 				if code == 104: # "Connection reset by peer."
 					continue
 			raise NetworkError(cause=exc)
-		except IOError, exc:
+		except socket.timeout as exc:
+			_log.debug("socket timeout")
+			last_exc = exc
+		except socket.error as exc:
+			if exc.errno == 104:
+				continue
+			raise NetworkError(cause=exc)
+		except IOError as exc:
 			raise NetworkError(cause=exc)
-		else:
-			# No exception! Yay!
-			return f
 
 	# Out of retries!
 	raise NetworkError("retried %i times" % max_retries, last_exc)
@@ -415,30 +465,30 @@ def _mb_request(path, method='GET', auth_required=False, client_required=False,
 
 	if _useragent == "":
 		raise UsageError("set a proper user-agent with "
-						 "musicbrainz.set_useragent(\"application name\", \"application version\", \"contact info (preferably URL or email for your application)\")")
+						 "set_useragent(\"application name\", \"application version\", \"contact info (preferably URL or email for your application)\")")
 
 	if client_required:
 		args["client"] = _client
 
 	# Encode Unicode arguments using UTF-8.
 	for key, value in args.items():
-		if isinstance(value, unicode):
+		if isinstance(value, compat.unicode):
 			args[key] = value.encode('utf8')
 
 	# Construct the full URL for the request, including hostname and
 	# query string.
-	url = urlparse.urlunparse((
+	url = compat.urlunparse((
 		'http',
 		hostname,
 		'/ws/2/%s' % path,
 		'',
-		urllib.urlencode(args),
+		compat.urlencode(args),
 		''
 	))
 	_log.debug("%s request for %s" % (method, url))
 
 	# Set up HTTP request handler and URL opener.
-	httpHandler = urllib2.HTTPHandler(debuglevel=0)
+	httpHandler = compat.HTTPHandler(debuglevel=0)
 	handlers = [httpHandler]
 
 	# Add credentials if required.
@@ -446,13 +496,13 @@ def _mb_request(path, method='GET', auth_required=False, client_required=False,
 		_log.debug("Auth required for %s" % url)
 		if not user:
 			raise UsageError("authorization required; "
-							 "use musicbrainz.auth(u, p) first")
+							 "use auth(user, pass) first")
 		passwordMgr = _RedirectPasswordMgr()
 		authHandler = _DigestAuthHandler(passwordMgr)
 		authHandler.add_password("musicbrainz.org", (), user, password)
 		handlers.append(authHandler)
 
-	opener = urllib2.build_opener(*handlers)
+	opener = compat.build_opener(*handlers)
 
 	# Make request.
 	req = _MusicbrainzHttpRequest(method, url, data)
@@ -460,15 +510,22 @@ def _mb_request(path, method='GET', auth_required=False, client_required=False,
 	_log.debug("requesting with UA %s" % _useragent)
 	if body:
 		req.add_header('Content-Type', 'application/xml; charset=UTF-8')
-	f = _safe_open(opener, req, body)
+	elif not data and not req.has_header('Content-Length'):
+		# Explicitly indicate zero content length if no request data
+		# will be sent (avoids HTTP 411 error).
+		req.add_header('Content-Length', '0')
+	resp = _safe_read(opener, req, body)
 
 	# Parse the response.
 	try:
-		return mbxml.parse_message(f)
-	except etree.ParseError, exc:
-		raise ResponseError(cause=exc)
-	except UnicodeError, exc:
+		return mbxml.parse_message(resp)
+	except UnicodeError as exc:
 		raise ResponseError(cause=exc)
+	except Exception as exc:
+		if isinstance(exc, ETREE_EXCEPTIONS):
+			raise ResponseError(cause=exc)
+		else:
+			raise
 
 def _is_auth_required(entity, includes):
 	""" Some calls require authentication. This returns
@@ -490,6 +547,8 @@ def _do_mb_query(entity, id, includes=[], params={}):
 	response is parsed and returned.
 	"""
 	# Build arguments.
+	if not isinstance(includes, list):
+		includes = [includes]
 	_check_includes(entity, includes)
 	auth_required = _is_auth_required(entity, includes)
 	args = dict(params)
@@ -501,15 +560,28 @@ def _do_mb_query(entity, id, includes=[], params={}):
 	path = '%s/%s' % (entity, id)
 	return _mb_request(path, 'GET', auth_required, args=args)
 
-def _do_mb_search(entity, query='', fields={}, limit=None, offset=None):
+def _do_mb_search(entity, query='', fields={},
+		  limit=None, offset=None, strict=False):
 	"""Perform a full-text search on the MusicBrainz search server.
-	`query` is a free-form query string and `fields` is a dictionary
+	`query` is a lucene query string when no fields are set,
+	but is escaped when any fields are given. `fields` is a dictionary
 	of key/value query parameters. They keys in `fields` must be valid
 	for the given entity type.
 	"""
 	# Encode the query terms as a Lucene query string.
-	query_parts = [query.replace('\x00', '').strip()]
-	for key, value in fields.iteritems():
+	query_parts = []
+	if query:
+		clean_query = util._unicode(query)
+		if fields:
+			clean_query = re.sub(r'([+\-&|!(){}\[\]\^"~*?:\\])',
+					r'\\\1', clean_query)
+			if strict:
+				query_parts.append('"%s"' % clean_query)
+			else:
+				query_parts.append(clean_query.lower())
+		else:
+			query_parts.append(clean_query)
+	for key, value in fields.items():
 		# Ensure this is a valid search field.
 		if key not in VALID_SEARCH_FIELDS[entity]:
 			raise InvalidSearchFieldError(
@@ -517,11 +589,19 @@ def _do_mb_search(entity, query='', fields={}, limit=None, offset=None):
 			)
 
 		# Escape Lucene's special characters.
+		value = util._unicode(value)
 		value = re.sub(r'([+\-&|!(){}\[\]\^"~*?:\\])', r'\\\1', value)
-		value = value.replace('\x00', '').strip()
 		if value:
-			query_parts.append(u'%s:(%s)' % (key, value))
-	full_query = u' '.join(query_parts).strip()
+			if strict:
+				query_parts.append('%s:"%s"' % (key, value))
+			else:
+				value = value.lower() # avoid AND / OR
+				query_parts.append('%s:(%s)' % (key, value))
+	if strict:
+		full_query = ' AND '.join(query_parts).strip()
+	else:
+		full_query = ' '.join(query_parts).strip()
+
 	if not full_query:
 		raise ValueError('at least one query term is required')
 
@@ -555,23 +635,23 @@ def _do_mb_post(path, body):
 
 # Single entity by ID
 def get_artist_by_id(id, includes=[], release_status=[], release_type=[]):
-	params = _check_filter_and_make_params(includes, release_status, release_type)
+	params = _check_filter_and_make_params("artist", includes, release_status, release_type)
 	return _do_mb_query("artist", id, includes, params)
 
 def get_label_by_id(id, includes=[], release_status=[], release_type=[]):
-	params = _check_filter_and_make_params(includes, release_status, release_type)
+	params = _check_filter_and_make_params("label", includes, release_status, release_type)
 	return _do_mb_query("label", id, includes, params)
 
 def get_recording_by_id(id, includes=[], release_status=[], release_type=[]):
-	params = _check_filter_and_make_params(includes, release_status, release_type)
+	params = _check_filter_and_make_params("recording", includes, release_status, release_type)
 	return _do_mb_query("recording", id, includes, params)
 
 def get_release_by_id(id, includes=[], release_status=[], release_type=[]):
-	params = _check_filter_and_make_params(includes, release_status, release_type)
+	params = _check_filter_and_make_params("release", includes, release_status, release_type)
 	return _do_mb_query("release", id, includes, params)
 
 def get_release_group_by_id(id, includes=[], release_status=[], release_type=[]):
-	params = _check_filter_and_make_params(includes, release_status, release_type)
+	params = _check_filter_and_make_params("release-group", includes, release_status, release_type)
 	return _do_mb_query("release-group", id, includes, params)
 
 def get_work_by_id(id, includes=[]):
@@ -579,60 +659,82 @@ def get_work_by_id(id, includes=[]):
 
 
 # Searching
+def search_annotations(query='', limit=None, offset=None, strict=False, **fields):
+	"""Search for annotations by a free-form `query` string or any of
+	the following keyword arguments specifying field queries:
+	entity, name, text, type
+	When `fields` are set, special lucene characters are escaped
+	in the `query`.
+	"""
+	return _do_mb_search('annotation', query, fields, limit, offset, strict)
 
-def search_artists(query='', limit=None, offset=None, **fields):
-	"""Search for artists by a free-form `query` string and/or any of
+def search_artists(query='', limit=None, offset=None, strict=False, **fields):
+	"""Search for artists by a free-form `query` string or any of
 	the following keyword arguments specifying field queries:
 	arid, artist, sortname, type, begin, end, comment, alias, country,
 	gender, tag
+	When `fields` are set, special lucene characters are escaped
+	in the `query`.
 	"""
-	return _do_mb_search('artist', query, fields, limit, offset)
+	return _do_mb_search('artist', query, fields, limit, offset, strict)
 
-def search_labels(query='', limit=None, offset=None, **fields):
-	"""Search for labels by a free-form `query` string and/or any of
+def search_labels(query='', limit=None, offset=None, strict=False, **fields):
+	"""Search for labels by a free-form `query` string or any of
 	the following keyword arguments specifying field queries:
 	laid, label, sortname, type, code, country, begin, end, comment,
 	alias, tag
+	When `fields` are set, special lucene characters are escaped
+	in the `query`.
 	"""
-	return _do_mb_search('label', query, fields, limit, offset)
+	return _do_mb_search('label', query, fields, limit, offset, strict)
 
-def search_recordings(query='', limit=None, offset=None, **fields):
-	"""Search for recordings by a free-form `query` string and/or any of
+def search_recordings(query='', limit=None, offset=None, strict=False, **fields):
+	"""Search for recordings by a free-form `query` string or any of
 	the following keyword arguments specifying field queries:
 	rid, recording, isrc, arid, artist, artistname, creditname, reid,
 	release, type, status, tracks, tracksrelease, dur, qdur, tnum,
 	position, tag
+	When `fields` are set, special lucene characters are escaped
+	in the `query`.
 	"""
-	return _do_mb_search('recording', query, fields, limit, offset)
+	return _do_mb_search('recording', query, fields, limit, offset, strict)
 
-def search_releases(query='', limit=None, offset=None, **fields):
-	"""Search for releases by a free-form `query` string and/or any of
+def search_releases(query='', limit=None, offset=None, strict=False, **fields):
+	"""Search for releases by a free-form `query` string or any of
 	the following keyword arguments specifying field queries:
 	reid, release, arid, artist, artistname, creditname, type, status,
 	tracks, tracksmedium, discids, discidsmedium, mediums, date, asin,
 	lang, script, country, date, label, catno, barcode, puid
+	When `fields` are set, special lucene characters are escaped
+	in the `query`.
 	"""
-	return _do_mb_search('release', query, fields, limit, offset)
+	return _do_mb_search('release', query, fields, limit, offset, strict)
 
-def search_release_groups(query='', limit=None, offset=None, **fields):
-	"""Search for release groups by a free-form `query` string and/or
+def search_release_groups(query='', limit=None, offset=None,
+			  strict=False, **fields):
+	"""Search for release groups by a free-form `query` string or
 	any of the following keyword arguments specifying field queries:
 	rgid, releasegroup, reid, release, arid, artist, artistname,
 	creditname, type, tag
+	When `fields` are set, special lucene characters are escaped
+	in the `query`.
 	"""
-	return _do_mb_search('release-group', query, fields, limit, offset)
+	return _do_mb_search('release-group', query, fields,
+			     limit, offset, strict)
 
-def search_works(query='', limit=None, offset=None, **fields):
-	"""Search for works by a free-form `query` string and/or any of
+def search_works(query='', limit=None, offset=None, strict=False, **fields):
+	"""Search for works by a free-form `query` string or any of
 	the following keyword arguments specifying field queries:
 	wid, work, iswc, type, arid, artist, alias, tag
+	When `fields` are set, special lucene characters are escaped
+	in the `query`.
 	"""
-	return _do_mb_search('work', query, fields, limit, offset)
+	return _do_mb_search('work', query, fields, limit, offset, strict)
 
 
 # Lists of entities
-def get_releases_by_discid(id, includes=[], release_type=[]):
-	params = _check_filter_and_make_params(includes, release_type=release_type)
+def get_releases_by_discid(id, includes=[], release_status=[], release_type=[]):
+	params = _check_filter_and_make_params(includes, release_status, release_type=release_type)
 	return _do_mb_query("discid", id, includes, params)
 
 def get_recordings_by_echoprint(echoprint, includes=[], release_status=[], release_type=[]):
@@ -669,26 +771,26 @@ def _browse_impl(entity, includes, valid_includes, limit, offset, params, releas
 # and the test in _do_mb_query will pass anyway.
 def browse_artists(recording=None, release=None, release_group=None, includes=[], limit=None, offset=None):
     # optional parameter work?
-    valid_includes = ["aliases", "tags", "ratings", "user-tags", "user-ratings"]
+    valid_includes = VALID_BROWSE_INCLUDES['artists']
     params = {"recording": recording,
               "release": release,
               "release-group": release_group}
     return _browse_impl("artist", includes, valid_includes, limit, offset, params)
 
 def browse_labels(release=None, includes=[], limit=None, offset=None):
-    valid_includes = ["aliases", "tags", "ratings", "user-tags", "user-ratings"]
+    valid_includes = VALID_BROWSE_INCLUDES['labels']
     params = {"release": release}
     return _browse_impl("label", includes, valid_includes, limit, offset, params)
 
 def browse_recordings(artist=None, release=None, includes=[], limit=None, offset=None):
-    valid_includes = ["artist-credits", "tags", "ratings", "user-tags", "user-ratings"]
+    valid_includes = VALID_BROWSE_INCLUDES['recordings']
     params = {"artist": artist,
               "release": release}
     return _browse_impl("recording", includes, valid_includes, limit, offset, params)
 
 def browse_releases(artist=None, label=None, recording=None, release_group=None, release_status=[], release_type=[], includes=[], limit=None, offset=None):
     # track_artist param doesn't work yet
-    valid_includes = ["artist-credits", "labels", "recordings", "release-groups"]
+    valid_includes = VALID_BROWSE_INCLUDES['releases']
     params = {"artist": artist,
               "label": label,
               "recording": recording,
@@ -696,7 +798,7 @@ def browse_releases(artist=None, label=None, recording=None, release_group=None,
     return _browse_impl("release", includes, valid_includes, limit, offset, params, release_status, release_type)
 
 def browse_release_groups(artist=None, release=None, release_type=[], includes=[], limit=None, offset=None):
-    valid_includes = ["artist-credits", "tags", "ratings", "user-tags", "user-ratings"]
+    valid_includes = VALID_BROWSE_INCLUDES['release-groups']
     params = {"artist": artist,
               "release": release}
     return _browse_impl("release-group", includes, valid_includes, limit, offset, params, [], release_type)
@@ -714,51 +816,65 @@ def get_releases_in_collection(collection):
 # Submission methods
 
 def submit_barcodes(barcodes):
-	"""
-	Submits a set of {release1: barcode1, release2:barcode2}
-	Must call auth(user, pass) first
-	"""
+	"""Submits a set of {release1: barcode1, release2:barcode2}
+
+	Must call auth(user, pass) first"""
 	query = mbxml.make_barcode_request(barcodes)
 	return _do_mb_post("release", query)
 
 def submit_puids(puids):
-	query = mbxml.make_puid_request(puids)
-	return _do_mb_post("recording", query)
+    """Submit PUIDs.
+
+    Must call auth(user, pass) first"""
+    query = mbxml.make_puid_request(puids)
+    return _do_mb_post("recording", query)
 
 def submit_echoprints(echoprints):
-	query = mbxml.make_echoprint_request(echoprints)
-	return _do_mb_post("recording", query)
+    """Submit echoprints.
+
+    Must call auth(user, pass) first"""
+    query = mbxml.make_echoprint_request(echoprints)
+    return _do_mb_post("recording", query)
 
 def submit_isrcs(recordings_isrcs):
-    """
-    Submit ISRCs.
-    Submits a set of {recording-id: [isrc1, irc1]}
-    Must call auth(user, pass) first
-    """
+    """Submit ISRCs.
+    Submits a set of {recording-id: [isrc1, isrc2, ...]}
+
+    Must call auth(user, pass) first"""
     query = mbxml.make_isrc_request(recordings_isrcs=recordings_isrcs)
     return _do_mb_post("recording", query)
 
 def submit_tags(artist_tags={}, recording_tags={}):
-	""" Submit user tags.
-	    Artist or recording parameters are of the form:
-	    {'entityid': [taglist]}
-	"""
-	query = mbxml.make_tag_request(artist_tags, recording_tags)
-	return _do_mb_post("tag", query)
+    """Submit user tags.
+    Artist or recording parameters are of the form:
+    {'entityid': [taglist]}
+
+    Must call auth(user, pass) first"""
+    query = mbxml.make_tag_request(artist_tags, recording_tags)
+    return _do_mb_post("tag", query)
 
 def submit_ratings(artist_ratings={}, recording_ratings={}):
-	""" Submit user ratings.
-	    Artist or recording parameters are of the form:
-	    {'entityid': rating}
-	"""
-	query = mbxml.make_rating_request(artist_ratings, recording_ratings)
-	return _do_mb_post("rating", query)
+    """ Submit user ratings.
+    Artist or recording parameters are of the form:
+    {'entityid': rating}
+
+    Must call auth(user, pass) first"""
+    query = mbxml.make_rating_request(artist_ratings, recording_ratings)
+    return _do_mb_post("rating", query)
 
 def add_releases_to_collection(collection, releases=[]):
-	# XXX: Maximum URI length of 16kb means we should only allow ~400 releases
-	releaselist = ";".join(releases)
-   	_do_mb_put("collection/%s/releases/%s" % (collection, releaselist))
+    """Add releases to a collection.
+    Collection and releases should be identified by their MBIDs
+
+    Must call auth(user, pass) first"""
+    # XXX: Maximum URI length of 16kb means we should only allow ~400 releases
+    releaselist = ";".join(releases)
+    _do_mb_put("collection/%s/releases/%s" % (collection, releaselist))
 
 def remove_releases_from_collection(collection, releases=[]):
-	releaselist = ";".join(releases)
-   	_do_mb_delete("collection/%s/releases/%s" % (collection, releaselist))
+    """Remove releases from a collection.
+    Collection and releases should be identified by their MBIDs
+
+    Must call auth(user, pass) first"""
+    releaselist = ";".join(releases)
+    _do_mb_delete("collection/%s/releases/%s" % (collection, releaselist))
